package main_test

import (
	"bytes"
	"context"
	"crypto/ecdsa"
	"errors"
	"fmt"
	"sync"

	//"encoding/hex"
	"math/big"
	"net"
	"testing"
	"time"

	"google.golang.org/grpc"
	"google.golang.org/grpc/credentials/insecure"
	"google.golang.org/protobuf/types/known/anypb"

	base "github.com/CaduceusMetaverseProtocol/metaprotocol/gen/proto/go/base/v1"
	ring "github.com/CaduceusMetaverseProtocol/metaprotocol/gen/proto/go/ring/v1"
	"github.com/ethereum/go-ethereum/common"
	"github.com/ethereum/go-ethereum/core/types"
	"github.com/ethereum/go-ethereum/crypto"
)

var countParallelAsEth int64
var countParallelAsStd int64
var countParallelAsAny int64
var countParallelAsBytes int64

type RingServer struct {
	ring.UnimplementedRingServiceServer
}

func (*RingServer) SendTxAsEth(ctx context.Context, in *base.TransactionEth) (*ring.SendRawTransactionResponse, error) {

	addr := common.Address{}

	copy(addr[:], in.Tx.Inner.Recipient.Address[:common.AddressLength])

	ethTx := types.NewTx(&types.LegacyTx{
		Nonce:    in.Tx.Inner.AccountNonce,
		To:       &addr,
		Value:    new(big.Int).SetBytes(in.Tx.Inner.Amount),
		Gas:      in.Tx.Inner.GasLimit,
		GasPrice: new(big.Int).SetBytes(in.Tx.Inner.Price),
		Data:     in.Tx.Inner.Payload,
		V:        new(big.Int).SetBytes(in.Tx.Inner.V),
		R:        new(big.Int).SetBytes(in.Tx.Inner.R),
		S:        new(big.Int).SetBytes(in.Tx.Inner.S),
	})

	// atomic.AddInt64(&countParallelAsEth, 1)
	// defer fmt.Println("countParallelAsEth", countParallelAsEth)

	return &ring.SendRawTransactionResponse{TxHash: ethTx.Hash().Bytes()}, nil
}

func (*RingServer) SendTxAsStd(ctx context.Context, in *base.TransactionStd) (*ring.SendRawTransactionResponse, error) {

	addr := common.Address{}

	copy(addr[:], in.Tx.Inner.Recipient.Address[:common.AddressLength])

	ethTx := types.NewTx(&types.LegacyTx{
		Nonce:    in.Tx.Inner.AccountNonce,
		To:       &addr,
		Value:    new(big.Int).SetBytes(in.Tx.Inner.Amount),
		Gas:      in.Tx.Inner.GasLimit,
		GasPrice: new(big.Int).SetBytes(in.Tx.Inner.Price),
		Data:     in.Tx.Inner.Payload,
		V:        new(big.Int).SetBytes(in.Tx.Inner.V),
		R:        new(big.Int).SetBytes(in.Tx.Inner.R),
		S:        new(big.Int).SetBytes(in.Tx.Inner.S),
	})

	// atomic.AddInt64(&countParallelAsStd, 1)

	// defer fmt.Println("countParallelAsStd", countParallelAsStd)

	return &ring.SendRawTransactionResponse{TxHash: ethTx.Hash().Bytes()}, nil
}

func (*RingServer) SendTxAsAny(ctx context.Context, in *base.Transaction) (*ring.SendRawTransactionResponse, error) {

	msg, err := in.Tx.UnmarshalNew()

	if err != nil {
		return nil, err
	}

	// atomic.AddInt64(&countParallelAsAny, 1)
	// defer fmt.Println("countParallelAsAny", countParallelAsAny)

	switch m := msg.(type) {

	case *base.EthTx:

		addr := common.Address{}

		copy(addr[:], m.Inner.Recipient.Address[:common.AddressLength])

		ethTx := types.NewTx(&types.LegacyTx{
			Nonce:    m.Inner.AccountNonce,
			To:       &addr,
			Value:    new(big.Int).SetBytes(m.Inner.Amount),
			Gas:      m.Inner.GasLimit,
			GasPrice: new(big.Int).SetBytes(m.Inner.Price),
			Data:     m.Inner.Payload,
			V:        new(big.Int).SetBytes(m.Inner.V),
			R:        new(big.Int).SetBytes(m.Inner.R),
			S:        new(big.Int).SetBytes(m.Inner.S),
		})

		return &ring.SendRawTransactionResponse{TxHash: ethTx.Hash().Bytes()}, nil

	case *base.StdTx:
		addr := common.Address{}

		copy(addr[:], m.Inner.Recipient.Address[:common.AddressLength])

		ethTx := types.NewTx(&types.LegacyTx{
			Nonce:    m.Inner.AccountNonce,
			To:       &addr,
			Value:    new(big.Int).SetBytes(m.Inner.Amount),
			Gas:      m.Inner.GasLimit,
			GasPrice: new(big.Int).SetBytes(m.Inner.Price),
			Data:     m.Inner.Payload,
			V:        new(big.Int).SetBytes(m.Inner.V),
			R:        new(big.Int).SetBytes(m.Inner.R),
			S:        new(big.Int).SetBytes(m.Inner.S),
		})

		return &ring.SendRawTransactionResponse{TxHash: ethTx.Hash().Bytes()}, nil

	default:
		return nil, errors.New(fmt.Sprintf("server expected tx type: %T and %T, but actually is %T", base.EthTx{}, base.StdTx{}, msg))
	}

}

func (*RingServer) SendTxAsBytes(ctx context.Context, in *base.TransactionBytes) (*ring.SendRawTransactionResponse, error) {

	ethTx := types.Transaction{}

	if err := ethTx.UnmarshalBinary(in.Tx); err != nil {
		return nil, err
	}

	// atomic.AddInt64(&countParallelAsBytes, 1)
	// defer fmt.Printf("countParallelAsBytes: %d\n", countParallelAsBytes)

	return &ring.SendRawTransactionResponse{TxHash: ethTx.Hash().Bytes()}, nil
}

func TestGrpcServer(t *testing.T) {

	lis, err := net.Listen("tcp", ":9006")
	if err != nil {
		t.Fatal(err)
	}

	s := grpc.NewServer()

	ring.RegisterRingServiceServer(s, &RingServer{})

	err = s.Serve(lis)
	if err != nil {
		t.Fatal(err)
	}
}

func pricedTransaction(to common.Address, nonce uint64, gaslimit uint64, gasprice *big.Int, key *ecdsa.PrivateKey) *types.Transaction {
	tx, _ := types.SignTx(types.NewTransaction(nonce, to, big.NewInt(100), gaslimit, gasprice, nil), types.HomesteadSigner{}, key)
	return tx
}

// go  test -v -run EthTx  -bench=.  -benchtime=3s
// go test   -v -run  BenchmarkEthTx -bench BenchmarkEthTx -benchtime=3s
// go test   -v -run  BenchmarkStdTx -bench BenchmarkStdTx -benchtime=3s
// go test   -v -run  BenchmarkAnyTx -bench BenchmarkAnyTx -benchtime=3s
// go test   -v -run  BenchmarkBytesEth -bench BenchmarkBytesEth -benchtime=3s

// go test   -v -run  BenchmarkEthTx -bench BenchmarkEthTx -benchtime=1s
// go test   -v -run  BenchmarkStdTx -bench BenchmarkStdTx -benchtime=1s
// go test   -v -run  BenchmarkAnyTx -bench BenchmarkAnyTx -benchtime=1s
// go test   -v -run  BenchmarkBytesEth -bench BenchmarkBytesEth -benchtime=1s

// go test   -v -run  BenchmarkEthTx -bench BenchmarkEthTx -benchtime=1x
// go test   -v -run  BenchmarkStdTx -bench BenchmarkStdTx -benchtime=1x
// go test   -v -run  BenchmarkAnyTx -bench BenchmarkAnyTx -benchtime=1x
// go test   -v -run  BenchmarkBytesEth -bench BenchmarkBytesEth -benchtime=1x

// go test   -v -run  BenchmarkEthTx -bench BenchmarkEthTx -benchtime=100x
// go test   -v -run  BenchmarkStdTx -bench BenchmarkStdTx -benchtime=100x
// go test   -v -run  BenchmarkAnyTx -bench BenchmarkAnyTx -benchtime=100x
// go test   -v -run  BenchmarkBytesEth -bench BenchmarkBytesEth -benchtime=100x

// go test   -v -run  BenchmarkEthTx -bench BenchmarkEthTx -benchtime=100x -benchmem
// go test   -v -run  BenchmarkStdTx -bench BenchmarkStdTx -benchtime=100x -benchmem
// go test   -v -run  BenchmarkAnyTx -bench BenchmarkAnyTx -benchtime=100x -benchmem
// go test   -v -run  BenchmarkBytesEth -bench BenchmarkBytesEth -benchtime=100x -benchmem

// go test -v -run TestGrpcServer  -timeout 0

//BenchmarkAny
// go  test -v -run Any  -bench=.  -benchtime=3s

var once sync.Once

var tx *types.Transaction

var fromAddr common.Address

// var local, remote *ecdsa.PrivateKey

func int() {

	local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
	publicKey := local.Public()
	publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
	fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

	remote, _ := crypto.GenerateKey()
	tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)

}

var count int64
var countParallel int64

var onceHash sync.Once

func BenchmarkEthTx(b *testing.B) {

	b.Logf("b.N: %d\n", b.N)

	onceFunc := func() {

		local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
		publicKey := local.Public()
		publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
		fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

		remote, _ := crypto.GenerateKey()
		tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)
	}

	once.Do(onceFunc)

	b.ResetTimer()

	b.RunParallel(func(pb *testing.PB) {

		for pb.Next() {
			conn, err := grpc.Dial("127.0.0.1:9006", grpc.WithTransportCredentials(insecure.NewCredentials()))
			if err != nil {
				b.Fatal(err)
			}
			defer conn.Close()

			c := ring.NewRingServiceClient(conn)

			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()

			inner := base.EthTxData{
				AccountNonce: tx.Nonce(),
				Price:        tx.GasPrice().Bytes(),
				GasLimit:     tx.Gas(),
				Payload:      tx.Data(),
			}

			v, r, sigs := tx.RawSignatureValues()

			inner.V = v.Bytes()
			inner.R = r.Bytes()
			inner.S = sigs.Bytes()
			inner.Amount = tx.Value().Bytes()

			addr := base.Address{Address: tx.To().Bytes()}

			inner.Recipient = &addr
			inner.From = fromAddr.Bytes()

			res, err := c.SendTxAsEth(ctx, &base.TransactionEth{Tx: &base.EthTx{Inner: &inner}})

			if err != nil {
				b.Fatal(err)
			}

			_ = res

			if bytes.Compare(tx.Hash().Bytes(), res.TxHash) != 0 {
				b.Fatal(err)
			}

			onceHash.Do(func() {
				b.Logf("response: %x  local: %x \n", res.TxHash, tx.Hash().Bytes())
			})
		}
	})

}

func BenchmarkStdTx(b *testing.B) {

	b.Logf("b.N: %d\n", b.N)

	onceFunc := func() {
		local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
		publicKey := local.Public()
		publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
		fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

		remote, _ := crypto.GenerateKey()
		tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)
	}

	once.Do(onceFunc)

	b.ResetTimer()

	b.RunParallel(func(pb *testing.PB) {

		for pb.Next() {

			conn, err := grpc.Dial("127.0.0.1:9006", grpc.WithTransportCredentials(insecure.NewCredentials()))
			if err != nil {
				b.Fatal(err)
			}
			defer conn.Close()

			c := ring.NewRingServiceClient(conn)

			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()

			inner := base.StdTxData{
				AccountNonce: tx.Nonce(),
				Price:        tx.GasPrice().Bytes(),
				GasLimit:     tx.Gas(),
				Payload:      tx.Data(),
			}

			v, r, sigs := tx.RawSignatureValues()

			inner.V = v.Bytes()
			inner.R = r.Bytes()
			inner.S = sigs.Bytes()
			inner.Amount = tx.Value().Bytes()

			addr := base.Address{Address: tx.To().Bytes()}

			inner.Recipient = &addr
			inner.From = fromAddr.Bytes()

			res, err := c.SendTxAsStd(ctx, &base.TransactionStd{Tx: &base.StdTx{Inner: &inner}})

			if err != nil {
				b.Fatal(err)
			}

			_ = res
			//fmt.Printf("%x \n", res.TxHash)
			if bytes.Compare(tx.Hash().Bytes(), res.TxHash) != 0 {
				b.Fatal(err)
			}

			onceHash.Do(func() {
				b.Logf("response: %x  local: %x \n", res.TxHash, tx.Hash().Bytes())
			})

		}
	})
}

func BenchmarkAnyTx(b *testing.B) {

	onceFunc := func() {
		local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
		publicKey := local.Public()
		publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
		fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

		remote, _ := crypto.GenerateKey()
		tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)
	}

	once.Do(onceFunc)

	b.ResetTimer()

	// The loop body is executed b.N times total across all goroutines.

	b.RunParallel(func(pb *testing.PB) {

		for pb.Next() {

			conn, err := grpc.Dial("127.0.0.1:9006", grpc.WithTransportCredentials(insecure.NewCredentials()))
			if err != nil {
				b.Fatal(err)
			}
			defer conn.Close()

			c := ring.NewRingServiceClient(conn)

			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()

			inner := base.EthTxData{
				AccountNonce: tx.Nonce(),
				Price:        tx.GasPrice().Bytes(),
				GasLimit:     tx.Gas(),
				Payload:      tx.Data(),
			}

			v, r, sigs := tx.RawSignatureValues()

			inner.V = v.Bytes()
			inner.R = r.Bytes()
			inner.S = sigs.Bytes()
			inner.Amount = tx.Value().Bytes()

			addr := base.Address{Address: tx.To().Bytes()}

			inner.Recipient = &addr

			ethTx := base.EthTx{Inner: &inner}

			ethTxAsAny, err := anypb.New(&ethTx)

			res, err := c.SendTxAsAny(ctx, &base.Transaction{Tx: ethTxAsAny})

			if err != nil {
				b.Fatal(err)
			}

			_ = res

			if bytes.Compare(tx.Hash().Bytes(), res.TxHash) != 0 {
				b.Fatal(err)
			}

			onceHash.Do(func() {
				b.Logf("response: %x  local: %x \n", res.TxHash, tx.Hash().Bytes())
			})

		}
	})
}

func BenchmarkBytesEth(b *testing.B) {

	onceFunc := func() {
		local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
		publicKey := local.Public()
		publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
		fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

		remote, _ := crypto.GenerateKey()
		tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)
	}

	once.Do(onceFunc)

	b.ResetTimer()

	b.RunParallel(func(pb *testing.PB) {

		for pb.Next() {

			conn, err := grpc.Dial("127.0.0.1:9006", grpc.WithTransportCredentials(insecure.NewCredentials()))
			if err != nil {
				b.Fatal(err)
			}
			defer conn.Close()

			c := ring.NewRingServiceClient(conn)

			ctx, cancel := context.WithTimeout(context.Background(), time.Second)
			defer cancel()

			txAsBytes, err := tx.MarshalBinary()
			// inner := base.EthTxData{
			// 	AccountNonce: tx.Nonce(),
			// 	Price:        tx.GasPrice().Bytes(),
			// 	GasLimit:     tx.Gas(),
			// 	Payload:      tx.Data(),
			// }

			// v, r, sigs := tx.RawSignatureValues()

			// inner.V = v.Bytes()
			// inner.R = r.Bytes()
			// inner.S = sigs.Bytes()
			// inner.Amount = tx.Value().Bytes()

			// addr := base.Address{Address: tx.To().Bytes()}

			// inner.Recipient = &addr
			// inner.From = fromAddr.Bytes()

			// txAsRes := &base.TransactionEth{Tx: &base.EthTx{Inner: &inner}}

			// txAsRes.

			// if err != nil {
			// 	b.Fatal(err)
			// }

			res, err := c.SendTxAsBytes(ctx, &base.TransactionBytes{Tx: txAsBytes, From: fromAddr[:]})

			if err != nil {
				b.Fatal(err)
			}

			_ = res

			if bytes.Compare(tx.Hash().Bytes(), res.TxHash) != 0 {
				b.Fatal(err)
			}

			onceHash.Do(func() {
				b.Logf("response: %x  local: %x \n", res.TxHash, tx.Hash().Bytes())
			})

		}
	})
}

func TestType(t *testing.T) {

	ethTx := base.EthTx{}

	fmt.Println(ethTx.ProtoReflect().Descriptor().FullName())

}

func TestAnyTx(t *testing.T) {

	onceFunc := func() {
		local, _ := crypto.HexToECDSA("FD5CC6F5E7E2805E920AC5DC83D5AF1106F9C92F0C04F9D5E1FD4261B4B4464A")
		publicKey := local.Public()
		publicKeyECDSA, _ := publicKey.(*ecdsa.PublicKey)
		fromAddr = crypto.PubkeyToAddress(*publicKeyECDSA)

		remote, _ := crypto.GenerateKey()
		tx = pricedTransaction(crypto.PubkeyToAddress(remote.PublicKey), 0, 100000, big.NewInt(1), local)
	}

	once.Do(onceFunc)

	conn, err := grpc.Dial("127.0.0.1:9006", grpc.WithTransportCredentials(insecure.NewCredentials()))
	if err != nil {
		t.Fatal(err)
	}
	defer conn.Close()

	c := ring.NewRingServiceClient(conn)

	ctx, cancel := context.WithTimeout(context.Background(), time.Second)
	defer cancel()

	_, _ = c, ctx

	inner := base.EthTxData{
		AccountNonce: tx.Nonce(),
		Price:        tx.GasPrice().Bytes(),
		GasLimit:     tx.Gas(),
		Payload:      tx.Data(),
	}

	v, r, sigs := tx.RawSignatureValues()

	inner.V = v.Bytes()
	inner.R = r.Bytes()
	inner.S = sigs.Bytes()
	inner.Amount = tx.Value().Bytes()

	addr := base.Address{Address: tx.To().Bytes()}

	inner.Recipient = &addr

	ethTx := base.EthTx{Inner: &inner}

	ethTxAsAny, err := anypb.New(&ethTx)

	if err != nil {
		t.Fatal(err)
	}

	//fmt.Println("ethTxAsAny.ProtoReflect().Descriptor().FullName()", ethTx.ProtoReflect().Descriptor().FullName())
	res, err := c.SendTxAsAny(ctx, &base.Transaction{Tx: ethTxAsAny})

	_ = res
	if err != nil {
		t.Fatal(err)
	}

	stdInner := base.StdTxData{
		AccountNonce: tx.Nonce(),
		Price:        tx.GasPrice().Bytes(),
		GasLimit:     tx.Gas(),
		Payload:      tx.Data(),
	}

	// v, r, sigs := tx.RawSignatureValues()

	stdInner.V = v.Bytes()
	stdInner.R = r.Bytes()
	stdInner.S = sigs.Bytes()
	stdInner.Amount = tx.Value().Bytes()

	//addr := base.Address{Address: tx.To().Bytes()}

	stdInner.Recipient = &addr

	stdTx := base.StdTx{Inner: &stdInner}

	_ = stdTx

	//stdTxAsAny, err := pbany(stdTx.ProtoReflect())

	stdTxAsAny, err := anypb.New(&stdTx)

	if err != nil {
		t.Fatal(err)
	}

	_ = stdTxAsAny

	res, err = c.SendTxAsAny(ctx, &base.Transaction{Tx: stdTxAsAny})

	if err != nil {
		t.Fatal(err)
	}

	_ = res

}

//go test   -v -run  BenchmarkHello -bench BenchmarkHello -benchtime=3s

func BenchmarkHello(b *testing.B) {
	b.Log(b.N)
	for i := 0; i < b.N; i++ {
		fmt.Sprintf("hello")
	}
}
